# This file is part of Octopus Sensing Processing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © 2022 Nastaran Saffaryazdi <nsaffar@gmail.com>
#
# Octopus Sensing Processing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing Processing is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing
# Processing. If not, see <https://www.gnu.org/licenses/>.

import threading
import json
import cherrypy

class ResultEndpoint:

    def __init__(self):
        self._result = None

    def set_result(self, result):
        self._result = result

    @cherrypy.expose
    def get_processing_result(self):
        return json.dumps(self._result)


class ResultEndpointServer(threading.Thread):
    def __init__(self, host: str="0.0.0.0", port: int=9333):
        super().__init__()
        self._host = host
        self._port = port

    def run(self):
        self._endpoint = ResultEndpoint()

        cherrypy.tree.mount(self._endpoint, '/')

        cherrypy.server.socket_host = self._host
        cherrypy.server.socket_port = self._port
        cherrypy.engine.autoreload.on = False
        cherrypy.engine.start()
        cherrypy.engine.block()

    def set_result(self, result):
        self._endpoint.set_result(result)



# This file is part of Octopus Sensing Processing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © 2022 Nastaran Saffaryazdi <nsaffar@gmail.com>
#
# Octopus Sensing Processing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing Processing is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing
# Processing. If not, see <https://www.gnu.org/licenses/>.

from operator import imod
import cherrypy
import http.client

class ResultEndpoint:

    def __init__(self):
        self._result = None

    @cherrypy.expose
    @cherrypy.tools.json_in()
    def index(self):
        print(cherrypy.request.json)


def test_streaming():
    endpoint = ResultEndpoint()

    cherrypy.tree.mount(endpoint, '/')

    cherrypy.server.socket_host = '0.0.0.0'
    cherrypy.server.socket_port = 9332
    cherrypy.engine.autoreload.on = False
    cherrypy.engine.start()
    cherrypy.engine.block()

def test_endpoint():
    connection = http.client.HTTPConnection('localhost', 9333, timeout=10)
    print(connection)
    while True:
        command = input('Press any key if you want to request a processing result. Otherwise press "q" to exit')
        if command == "q":
            break
        connection.request("GET", "/get_processing_result")
        response = connection.getresponse()
        print("Status: {} and reason: {}".format(response.status, response.reason))
        print(response.read())

    connection.close()

#test_streaming()
test_endpoint()



# This file is part of Octopus Sensing Processing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © 2022 Nastaran Saffaryazdi <nsaffar@gmail.com>
#
# Octopus Sensing Processing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing Processing is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing
# Processing. If not, see <https://www.gnu.org/licenses/>.

import sys
import time
import argparse
import importlib

from octopus_sensing_processing.octopus_sensing_client import OctopusSensingClient

from octopus_sensing_processing.result_endpoint import ResultEndpointServer
from octopus_sensing_processing.result_streaming import ResultStreaming

def get_args():
    arg_parser = argparse.ArgumentParser()

    arg_parser.add_argument("-t", "--time",
                            help="The time duration of getting data in seconds.",
                            default=3,
                            type=int)

    arg_parser.add_argument("-d", "--devices",
                            help="The list of devices names (The names passed to device coordinator). It can be like this: device1,device2",
                            type=str,
                            required=True)
    arg_parser.add_argument("-s", "--streaming",
                            help="Stream the result. Example: http://127.0.0.1:9332/",
                            type=str)
    arg_parser.add_argument("-e", "--endpoint",
                            help="GET endpoint for the result. Example: 0.0.0.0:9333",
                            type=str)
    arg_parser.add_argument("-m", "--module",
                            help="The module containing the processor function. Example: /path/to/processor_module.py",
                            required=True)
    arg_parser.add_argument("-f", "--function",
                            help="The name of the processor function. Example: process_data",
                            required=True)
    arg_parser.add_argument("-r", "--realtime_data_endpoint",
                            help="The realtime data for processing. Example: http://127.0.0.1:9330/",
                            type=str,
                            default="localhost:9330")

    return arg_parser.parse_args()


def main():
    args = get_args()

    # Dynamically import the processor module and get the process_data function
    module_name = "process_data"
 
    spec = importlib.util.spec_from_file_location(module_name, args.module)
    processor_module = importlib.util.module_from_spec(spec)

    # Optional but recommended: register in sys.modules
    sys.modules[module_name] = processor_module

    spec.loader.exec_module(processor_module)

    processor = getattr(processor_module, args.function)

    data_server_address = args.realtime_data_endpoint.split(":")
    client = OctopusSensingClient(host=data_server_address[0], port=data_server_address[1])

    if args.endpoint is not None:
        server_address = args.endpoint.split(":")
        if len(server_address) < 2:
            raise RuntimeError("The endpoint addrees is not in the correct format. It should be something like this: 0.0.0.0:9333")
        result_endpoint = ResultEndpointServer(host=server_address[0], port=int(server_address[1]))
        result_endpoint.start()

    if args.streaming is not None:
        result_streaming = ResultStreaming(url=args.streaming)
    while True:
        start = time.time()

        data = client.fetch(duration=args.time, device_list=args.devices)
        processing_result = processor(data)
        print(processing_result)
        print("octopus", processing_result)
        if args.streaming is not None:
            result_streaming.push_processing_result(processing_result)
        if args.endpoint is not None:
            result_endpoint.set_result(processing_result)

        end = time.time()
        computation_delay = end-start
        if computation_delay < args.time:
            time.sleep(args.time - computation_delay)

if __name__ == '__main__':
    main()



# This file is part of Octopus Sensing Processing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © 2022 Nastaran Saffaryazdi <nsaffar@gmail.com>
#
# Octopus Sensing Processing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing Processing is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing
# Processing. If not, see <https://www.gnu.org/licenses/>.

import json
import urllib.request
import logging
import time

# The data received from the processor might is in numpy format.
# We need to import 'numpy' to unpickle it.

class ResultStreaming:

    def __init__(self, url: str= "http://127.0.0.1:9332/"):

        self._url = url

    def push_processing_result(self, result):
        result_json = json.dumps(result).encode('utf-8')

        self.__post_data(self._url, result_json, retries=60)
        
    def __post_data(self, result_json, retries=3):
            try:
                request = urllib.request.Request(self._url,
                                                headers={"Content-Type": "application/json"},
                                                method='POST',
                                                data=result_json)
                with urllib.request.urlopen(request) as response:
                    if response.getcode() != 200:
                        raise RuntimeError("Posting data failed. Got HTTP status: {0} {1}".format(response.status,
                                                                                                  response.reason))

                    serialized_data = response.read()
                    return serialized_data
            except Exception as err:
                if retries == 0:
                    logging.error(f"Failed to post the data: {err}")
                    raise
                else:
                    logging.error(f"Posting data failed. Re-trying [{retries}]. Error: {err}")
                    time.sleep(1)
                    self.__post_data(self._url, result_json, retries=retries - 1)



# This file is part of Octopus Sensing Processing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © 2022 Nastaran Saffaryazdi <nsaffar@gmail.com>
#
# Octopus Sensing Processing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing Processing is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing
# Processing. If not, see <https://www.gnu.org/licenses/>.

import pickle
import urllib.request
import logging
import time
# The data received from the server might be in numpy format.
# We need to import 'numpy' to unpickle it.
from typing import Optional, List


class OctopusSensingClient:
    def __init__(self, host: str= "127.0.0.1", port: int=9330):
        self._host = host
        self._port = port

    def fetch(self, duration: int=3, device_list: str=None):
        if device_list is None:
            url = "http://{0}:{1}/?duration={2}".format(self._host, self._port, duration)
        else:
            # device_list can be like this "device1,device2"
            url = "http://{0}:{1}/?duration={2}&device_list={3}".format(self._host,
                                                                        self._port,
                                                                        duration,
                                                                        device_list)
        try:
            serialized_data = self.__request_data(url, retries=60)
        except Exception as error:
            raise error
        
        raw_data = pickle.loads(serialized_data)
        data = {}

        for key, device_data in raw_data.items():
            if device_data["metadata"]["type"] == "TestDeviceStreaming":
                data.setdefault("test", {})
                data["test"]["data"] = device_data["data"]
                data["test"]["sampling_rate"] = device_data["metadata"]["sampling_rate"]
                data["test"]["channels"] = device_data["metadata"]["channels"]
            if device_data["metadata"]["type"] == "OpenBCIStreaming":
                data.setdefault("eeg", {})
                data["eeg"]["data"] = self._restructure_eeg(device_data["data"])
                data["eeg"]["sampling_rate"] = device_data["metadata"]["sampling_rate"]
                data["eeg"]["channels"] = device_data["metadata"]["channels"]

            if device_data["metadata"]["type"] == "BrainFlowOpenBCIStreaming":
                data.setdefault("eeg", {})
                data["eeg"]["data"] = self._restructure_eeg(device_data["data"])
                data["eeg"]["sampling_rate"] = device_data["metadata"]["sampling_rate"]
                data["eeg"]["channels"] = device_data["metadata"]["channels"]

            if device_data["metadata"]["type"] == "Shimmer3Streaming":
                data.setdefault("ppg", {})
                data.setdefault("gsr", {})
                gsr_records, ppg_records, = self._restructure_shimmer(device_data["data"])
                data["gsr"]["data"] = gsr_records
                data["ppg"]["data"] = ppg_records
                data["gsr"]["sampling_rate"] = device_data["metadata"]["sampling_rate"]
                data["ppg"]["sampling_rate"] = device_data["metadata"]["sampling_rate"]

            if device_data["metadata"]["type"] == "CameraStreaming":
                data.setdefault("camera", {})
                # The type of data is a list of numpy-array. Each item is a frame
                data["camera"]["data"] = device_data["data"]
                data["camera"]["frame_rate"] = device_data["metadata"]["frame_rate"]

        return data

    def __request_data(self, url, retries=60):
            try:
                request = urllib.request.Request(url,
                                                headers={"Accept": "application/pickle"},
                                                method='GET')
                with urllib.request.urlopen(request) as response:
                    if response.getcode() != 200:
                        raise RuntimeError("Getting data failed. Got HTTP status: {0} {1}".format(response.status,
                                                                                                  response.reason))
                        
                    serialized_data = response.read()
                    return serialized_data
            except Exception as err:
                if retries == 0:
                    logging.error(f"Failed to get the data: {err}")
                    raise
                else:
                    logging.error(f"Getting data failed. Re-trying [{retries}]. Error: {err}")
                    time.sleep(1)
                    return self.__request_data(url, retries=retries - 1)


    def _restructure_eeg(self, data):
        channels = 8
        if len(data[0]) >= 16:
            channels = 16

        result = []

        for _ in range(channels):
            result.append([])

        for record in data:
            for i, channel_data in enumerate(record[1:channels+1]):
                result[i].append(channel_data)

        return result

    def _restructure_shimmer(self, data):
        gsr_records = []
        ppg_records = []

        for record in data:
            gsr_records.append(record[5])
            ppg_records.append(record[6])

        return (gsr_records, ppg_records)

